# -*- coding: utf-8 -*-
"""
Module to generate specific rules and otherwise configurations to allow communication between the
microservices
"""
from troposphere import (
    Ref, Sub, GetAtt, Select, Tags
)
from troposphere.ec2 import (
    EIP, SecurityGroup, SecurityGroupIngress
)
from troposphere.ecs import (
    LoadBalancer as EcsLoadBalancer
)
from troposphere.elasticloadbalancingv2 import (
    LoadBalancer,
    LoadBalancerAttributes,
    TargetGroup, TargetGroupAttribute,
    Listener,
    Action as ListenerAction,
    SubnetMapping
)

from ecs_composex.common import KEYISSET, LOG
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.ecs.ecs_networking_ingress import (
    add_lb_to_service_ingress,
    add_public_security_group_ingress
)
from ecs_composex.ecs.ecs_params import (
    SERVICE_NAME_T, SERVICE_NAME, SG_T
)
from ecs_composex.vpc.vpc_params import (
    VPC_ID, PUBLIC_SUBNETS
)


def define_grace_period(template, service):
    """Function to define grace period
    TO IMPLEMENT
    """
    return Ref('AWS::NoValue')


def add_public_ips(template, service_name, azs):
    """Function to add EIPs for each AZ selected for build

    :param template: template of the service to add the resources
    :type template: troposphere.Template
    :param service_name: name of the service
    :type service_name: str
    :param azs: list of AZs to deploy the EIPs to
    :type azs: list

    :return: list of troposphere.ec2.EIP
    :rtype: list
    """
    eips = []
    for az in azs:
        eips.append(
            EIP(
                f"EipPublicNlb{az.replace('-', '').strip()}{service_name}",
                template=template,
                Domain='vpc'
            )
        )
    return eips


def define_public_mapping(eips, azs):
    """Function to get the public mapping for NLB

    :param eips: list of EIPSs
    :type eips: list(troposphere.ec2.EIP)
    :param azs: list of AZs to created EIPs into
    :type azs: list

    :return: list
    """
    public_mappings = []
    if eips:
        public_mappings = [
            SubnetMapping(
                AllocationId=GetAtt(eip, 'AllocationId'),
                SubnetId=Select(count, Ref(PUBLIC_SUBNETS))
            ) for count, eip in enumerate(eips)
        ]
    elif azs:
        public_mappings = [
            SubnetMapping(
                SubnetId=Select(count, Ref(PUBLIC_SUBNETS))
            ) for count in range(len(azs))
        ]
    return public_mappings


def add_alb_sg(template, ports, public):
    """Function to add a security group for application loadbalancer

    :param template: template of the service to add resources to
    :type template: troposphere.Template
    :param ports: list of ports to add ingress from the ALB to service to
    :type ports: list of ports
    :param public: whether the ALB is public
    :type public: int

    :return: The ALB's SG
    :rtype: troposphere.ec2.SecurityGroup
    """
    suffix = 'Private'
    if public:
        suffix = "Public"
    sg = SecurityGroup(
        f"AlbSecurityGroup{suffix}",
        template=template,
        GroupDescription=Sub(f"ALB SG for ${{{SERVICE_NAME_T}}} in ${{{ROOT_STACK_NAME_T}}}"),
        VpcId=Ref(VPC_ID),
        Tags=Tags(
            {
                'Name': Sub(f"alb-sg-${{{SERVICE_NAME_T}}}-${{{ROOT_STACK_NAME_T}}}"),
                'StackName': Ref('AWS::StackName'),
                'MicroserviceName': Ref(SERVICE_NAME)
            }
        )
    )
    for port in ports:
        SecurityGroupIngress(
            f'FromAlbToServiceOnPort{port}',
            template=template,
            FromPort=port,
            ToPort=port,
            GroupId=GetAtt(SG_T, 'GroupId'),
            SourceSecurityGroupId=GetAtt(sg, 'GroupId'),
            SourceSecurityGroupOwnerId=Ref('AWS::AccountId'),
            IpProtocol='tcp'
        )
    return sg


def add_lb_listener(template, port, lb, settings, tgt):
    """Function add listener for the LB

    :param template: template of the service to add resources to
    :type template: troposphere.Template
    :param port: port to add the listener for
    :type port: int
    :param lb: the loadbalancer the listener depends on
    :type lb: tropopshere.elasticloadbalancingv2.LoadBalancer
    :param settings: network settings as defined in :func: ecs_composex.ecs.ecs_networking
    :type settings: dict
    :param tgt: the target group to associate
    :type tgt: troposphere.elasticloadbalancingv2.TargetGroup

    :return: listener
    :rtype: troposphere.elasticloadbalancingv2.Listener
    """
    suffix = 'Private'
    if settings['is_public']:
        suffix = 'Public'
    listener = Listener(
        f"{settings['lb_type'].title()}{suffix}ListenerPort{port}",
        template=template,
        DependsOn=[lb],
        DefaultActions=[
            ListenerAction(
                Type='forward',
                TargetGroupArn=Ref(tgt)
            )
        ],
        LoadBalancerArn=Ref(lb),
        Port=port,
        Protocol='TCP' if settings['lb_type'] == 'network' else 'HTTP'
    )
    return listener


def add_target_group(template, port, lb, settings):
    """Function to generate the TargetGroups

    :param template: the service template to add resources to
    :type template: troposphere.Template
    :param port: the port to add the targetgroup for
    :type port: int
    :param lb: the loadbalancer the targetgroup will be bound to
    :type lb: troposphere.elasticloadbalancingv2.LoadBalancer
    :param settings: network settings as defined in :func: ecs_composex.ecs.ecs_networking
    :type settings: dict

    :return: target group
    :rtype: troposphere.elasticloadbalancingv2.TargetGroup
    """
    suffix = 'Private'
    if settings['is_public']:
        suffix = 'Public'
    tgt = TargetGroup(
        f"{settings['lb_type'].title()}{suffix}TargetGroupPort{port}".strip(),
        template=template,
        DependsOn=[lb],
        VpcId=Ref(VPC_ID),
        Port=port,
        Protocol='TCP' if settings['lb_type'] == 'network' else 'HTTP',
        TargetType='ip',
        HealthCheckIntervalSeconds=10,
        HealthyThresholdCount=2,
        UnhealthyThresholdCount=2,
        TargetGroupAttributes=[
            TargetGroupAttribute(
                Key='deregistration_delay.timeout_seconds',
                Value='10'
            )
        ],
        Tags=Tags(
            {
                'Name': Sub(f"${{{SERVICE_NAME_T}}}-{port}"),
                'StackName': Ref('AWS::StackName'),
                'StackId': Ref('AWS::StackId'),
                'MicroserviceName': Ref(SERVICE_NAME_T)
            }
        )
    )
    return tgt


def add_load_balancer(template, service_name, settings, ports, **kwargs):
    """Function to add LB to template

    :param template: template to add the resources to
    :type template: troposphere.Template
    :param service_name: name of the service
    :type service_name: str
    :param settings: network settings as defined in :func: ecs_composex.ecs.ecs_networking
    :type settings: dict
    :param ports: list of ports to accept ingress for that service
    :type ports: list
    :param kwargs: optional arguments
    :type: dict

    :return: loadbalancer
    :rtype: troposphere.elasticloadbalancingv2.LoadBalancer
    """
    alb_sg = None
    eips = []
    if KEYISSET('is_public', settings) and settings['lb_type'] == 'network':
        eips = add_public_ips(template, service_name, kwargs['AwsAzs'])

    no_value = Ref('AWS::NoValue')
    public_mapping = define_public_mapping(eips, kwargs['AwsAzs'])
    if ports and settings['lb_type'] == 'application':
        alb_sg = add_alb_sg(template, ports, settings['is_public'])
        add_lb_to_service_ingress(template, alb_sg, SG_T, settings)
        lb_sg = [Ref(alb_sg)]
    else:
        lb_sg = no_value

    loadbalancer = LoadBalancer(
        f"Microservice{settings['lb_type'].title()}LB",
        template=template,
        Scheme='internet-facing' if settings['is_public'] else 'internal',
        LoadBalancerAttributes=[
            LoadBalancerAttributes(
                Key='load_balancing.cross_zone.enabled',
                Value='true'
            )
        ] if settings['lb_type'] == 'network' else no_value,
        SecurityGroups=lb_sg,
        SubnetMappings=public_mapping if settings['is_public'] and settings['lb_type'] == 'network' else no_value,
        Subnets=Ref(PUBLIC_SUBNETS) if settings['is_public'] and settings['lb_type'] == 'application' else no_value,
        Type=settings['lb_type'],
        Tags=Tags(
            {
                'Name': Sub(f"${{{SERVICE_NAME_T}}}-${{{ROOT_STACK_NAME_T}}}"),
                'StackName': Ref('AWS::StackName'),
                'MicroserviceName': Ref(SERVICE_NAME)
            }
        )
    )
    if settings['is_public']:
        if settings['lb_type'] == 'application' and alb_sg:
            add_public_security_group_ingress(
                template, service_name, settings, alb_sg
            )
        elif settings['lb_type'] == 'network':
            add_public_security_group_ingress(
                template, service_name, settings, SG_T
            )
    return loadbalancer


def add_service_load_balancer(template, service_name, settings, **kwargs):
    """Function to add all ELBv2 resources for a microservice

    :param template: template to add the resources to
    :type template: troposphere.Template
    :param service_name:
    :type service_name: str
    :param settings:
    :type settings: dict
    :param kwargs:
    :type kwargs: dict

    :return: service_lbs, depends_on
    :rtype: tuple
    """
    service_lbs = []
    tgt_groups = []
    depends_on = []
    curated_ports = [port['target'] for port in settings['ports']]
    service_lb = add_load_balancer(
        template, service_name, settings, curated_ports, **kwargs
    )
    depends_on.append(service_lb.title)
    for port in curated_ports:
        tgt = add_target_group(template, port, service_lb, settings)
        listener = add_lb_listener(template, port, service_lb, settings, tgt)
        tgt_groups.append(tgt)
        depends_on.append(tgt.title)
        depends_on.append(listener.title)

    for target in tgt_groups:
        service_lbs.append(
            EcsLoadBalancer(
                TargetGroupArn=Ref(target),
                ContainerPort=tgt.Port,
                ContainerName=Ref(SERVICE_NAME)
            )
        )
    return service_lbs, depends_on


def define_lb_type(service_name, labels):
    """Function to determine which LB is to be created

    :param service_name: name of the service
    :type service_name: str
    :param labels: labels to use to determine lb_type
    :type labels: dict

    :return: lb_type
    :rtype: str
    """
    lb_type = 'application'
    if KEYISSET('use_nlb', labels) and KEYISSET('use_alb', labels):
        LOG.warn('Both ALB and NLB are enabled for this service. Defaulting to ALB')
    elif KEYISSET('use_nlb', labels) and not KEYISSET('use_alb', labels):
        LOG.debug(f'Creating a NLB for service {service_name}')
        lb_type = 'network'
    elif not KEYISSET('use_nlb', labels) and KEYISSET('use_alb', labels):
        LOG.debug(f'Creating a ALB for service {service_name}')
    else:
        LOG.warn(
            'Neither ALB or NLB were specified but service was flagged as service.'
            'Defaulting to ALB'
        )
    return lb_type


def define_service_load_balancing(template, service_name, settings, **kwargs):
    """Function to add all the resources for when using ALB or NLB

    :param template: Service template to add resources to
    :type template: troposphere.Template
    :param service_name: name of the service
    :param settings: network settings defined in compile_network_settings
    :type settings: dict
    :param kwargs: Optional arguments
    :type kwargs: dict

    :return: service_lb
    :rtype: tuple
    """
    lb_type = define_lb_type(service_name, settings)
    settings.update({
        'lb_type': lb_type
    })
    LOG.debug(f'Adding LB for service {service_name}')
    return add_service_load_balancer(template, service_name, settings, **kwargs)



# -*- coding: utf-8 -*-
"""Module to generate a full stack with VPC & Cluster."""

import json

import boto3
from troposphere import Ref, GetAtt, If, Join
from troposphere.cloudformation import Stack
from troposphere.ecs import Cluster

from ecs_composex.compute import create_cluster_template
from ecs_composex.compute.cluster_params import (
    TARGET_CAPACITY_T,
    TARGET_CAPACITY,
    MIN_CAPACITY_T,
    USE_FLEET, USE_FLEET_T,
    USE_ONDEMAND, USE_ONDEMAND_T
)
from ecs_composex.ecs.ecs_params import CLUSTER_NAME_T, CLUSTER_NAME
from ecs_composex.ecs.ecs_conditions import (
    GENERATED_CLUSTER_NAME_CON_T,
    GENERATED_CLUSTER_NAME_CON,
    CLUSTER_NAME_CON_T,
    CLUSTER_NAME_CON
)
from ecs_composex.common import LOG, add_parameters
from ecs_composex.common import (
    build_template, KEYISSET,
    load_composex_file
)
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME_T
)
from ecs_composex.common.templates import upload_template
from ecs_composex.ecs import ecs_params, create_services_templates
from ecs_composex.ecs_composex import (
    RES_REGX, get_mod_function
)
from ecs_composex.vpc import create_vpc_template
from ecs_composex.vpc import vpc_params

ROOT_CLUSTER_NAME = 'EcsCluster'


def build_parameters_file(params: list, parameter_name: str, parameter_value):
    """
    Function to build arguments file to pass onto CFN.
    Adds the parameter key/value so it can be written to file afterwards

    :param params: list of parameters
    :type params: list
    :param parameter_name: key of the parameter
    :type parameter_name: str
    :param parameter_value: value of the parameter
    :type parameter_value: str or int or list
    """
    if params is None:
        params = []
    if isinstance(parameter_value, (int, float)):
        parameter_value = str(parameter_value)
    params.append({
        "ParameterKey": parameter_name,
        "ParameterValue": parameter_value
    })


def generate_vpc_parameters(template, params, **kwargs):
    """Function to add the VPC arguments to the root stack
    :param params: list of parameters
    :type params: list
    """
    build_parameters_file(
        params, vpc_params.VPC_ID_T, kwargs[vpc_params.VPC_ID_T]
    )
    build_parameters_file(
        params, vpc_params.APP_SUBNETS_T, ','.join(kwargs[vpc_params.APP_SUBNETS_T])
    )
    build_parameters_file(
        params, vpc_params.PUBLIC_SUBNETS_T, ','.join(kwargs[vpc_params.PUBLIC_SUBNETS_T])
    )
    build_parameters_file(
        params, vpc_params.STORAGE_SUBNETS_T, ','.join(kwargs[vpc_params.STORAGE_SUBNETS_T])
    )
    build_parameters_file(
        params, vpc_params.VPC_MAP_ID_T, kwargs[vpc_params.VPC_MAP_ID_T]
    )
    params = [
        vpc_params.VPC_ID,
        vpc_params.APP_SUBNETS,
        vpc_params.STORAGE_SUBNETS,
        vpc_params.PUBLIC_SUBNETS,
        vpc_params.VPC_MAP_ID
    ]
    add_parameters(template, params)


def add_vpc_to_root(root_template, session, kwargs):
    """
    Function to add VPC stack to the root one.

    :param root_template: root stack template
    :type root_template: troposphere.Template
    :param session: boto session for override
    :type session: boto3.session.Session
    :param kwargs:

    :return: vpc_stack
    :rtype: troposphere.cloudformation.Stack
    """
    vpc_template = create_vpc_template(
            session=session,
            **kwargs
        )
    vpc_template_url = upload_template(
        vpc_template.to_json(),
        kwargs['BucketName'],
        'vpc.json',
        session=session
    )
    LOG.debug(vpc_template_url)
    with open(f"{kwargs['output_file']}.vpc.json", 'w') as vpc_fd:
        vpc_fd.write(vpc_template.to_json())
    vpc_stack = root_template.add_resource(
        Stack(
            'Vpc',
            TemplateURL=vpc_template_url,
            Parameters={
                ROOT_STACK_NAME_T: Ref('AWS::StackName')
            }
        )
    )
    return vpc_stack


def add_cluster_to_root(root_template, params=None, vpc_stack=None, session=None, **kwargs):
    """
    Function to add Cluster stack to root one.

    :param root_template: the root template
    :type root_template: troposphere.Template
    :param params:
    :param vpc_stack:
    :param session:
    :param kwargs:

    :return:
    """
    if params is None:
        params = []
    depends_on = []
    root_template.add_parameter(TARGET_CAPACITY)
    cluster_template = create_cluster_template(session, **kwargs)
    parameters = {
        ROOT_STACK_NAME_T: Ref('AWS::StackName'),
        TARGET_CAPACITY_T: Ref(TARGET_CAPACITY),
        MIN_CAPACITY_T: Ref(TARGET_CAPACITY),
        USE_FLEET_T: Ref(USE_FLEET),
        USE_ONDEMAND_T: Ref(USE_ONDEMAND)
    }
    if vpc_stack is not None:
        depends_on.append(vpc_stack)
        parameters.update({
            vpc_params.VPC_ID_T: GetAtt(vpc_stack, f"Outputs.{vpc_params.VPC_ID_T}"),
            vpc_params.APP_SUBNETS_T: GetAtt(vpc_stack, f"Outputs.{vpc_params.APP_SUBNETS_T}"),
        })
    else:
        # Setup parameters for cluster stack without VPC pre-created
        if not kwargs['AppSubnets']:
            raise ValueError("No application subnets were provided to create the cluster")
        elif params is None:
            raise TypeError("params is ", params, "expected", list)
        root_template.add_parameter(vpc_params.APP_SUBNETS)
        parameters.update({
            vpc_params.APP_SUBNETS_T: Ref(vpc_params.APP_SUBNETS)
        })
        build_parameters_file(params, vpc_params.APP_SUBNETS_T, kwargs['AppSubnets'])

    cluster_stack = root_template.add_resource(Stack(
            'Compute',
            TemplateURL=upload_template(
                    template_body=cluster_template.to_json(),
                    bucket_name=kwargs['BucketName'],
                    file_name='cluster.json',
                    session=session
            ),
            Parameters=parameters
        )
    )
    return cluster_stack


def add_x_resources(template, session, **kwargs):
    """
    Function to add each X resource from the compose file
    """
    content = load_composex_file(kwargs['ComposeXFile'])
    depends_on = []
    ignore_x = ['x-rds']
    for key in content:
        if key.startswith('x-') and key not in ignore_x:
            res_type = RES_REGX.sub('', key)
            function_name = f"create_{res_type}_template"
            create_function = get_mod_function(res_type, function_name)
            if create_function:
                x_template = create_function(session=session, **kwargs)
                x_template_url = upload_template(
                    x_template.to_json(),
                    kwargs['BucketName'],
                    f"{res_type}.json",
                    session=session
                )
                depends_on.append(res_type.title().strip())
                template.add_resource(Stack(
                        res_type.title().strip(),
                        TemplateURL=x_template_url,
                        Parameters={
                            ROOT_STACK_NAME_T: Ref('AWS::StackName')
                        }
                    ))
    return depends_on


def add_services(template, depends, session, vpc_stack=None, **kwargs):
    """Function to add the microservices root stack

    :param cluster_stack: The cluster stack if created
    :type cluster_stack: troposphere.cloudformation.Stack
    :param template: root template
    :type template: troposphere.Template
    :param depends: list of dependencies for the stack
    :type depends: list
    :param session: ovveride boto session for API calls
    :type session: boto3.session.Session
    :param vpc_stack: the VPC Stack
    :type vpc_stack: troposphere.cloudformation.Template
    :param kwargs: optional parameters
    :type kwargs: dict or set
    """
    services_template = create_services_templates(session=session, **kwargs)
    services_template_url = upload_template(
        template_body=services_template.to_json(),
        bucket_name=kwargs['BucketName'],
        file_name='services.json'
    )
    parameters = {
        ROOT_STACK_NAME_T: Ref('AWS::StackName')
    }
    if not KEYISSET(CLUSTER_NAME_T, kwargs):
        parameters[ecs_params.CLUSTER_NAME_T] = Ref(ROOT_CLUSTER_NAME)
    else:
        parameters[ecs_params.CLUSTER_NAME_T] = Ref(CLUSTER_NAME)
    if vpc_stack:
        parameters.update({
            vpc_params.VPC_ID_T: GetAtt(vpc_stack, f'Outputs.{vpc_params.VPC_ID_T}'),
            vpc_params.PUBLIC_SUBNETS_T: GetAtt(vpc_stack, f'Outputs.{vpc_params.PUBLIC_SUBNETS_T}'),
            vpc_params.APP_SUBNETS_T: GetAtt(vpc_stack, f'Outputs.{vpc_params.APP_SUBNETS_T}'),

        })
    else:
        parameters.update({
            vpc_params.VPC_ID_T: Ref(vpc_params.VPC_ID),
            vpc_params.PUBLIC_SUBNETS_T: Join(',', Ref(vpc_params.PUBLIC_SUBNETS)),
            vpc_params.APP_SUBNETS_T: Join(',', Ref(vpc_params.APP_SUBNETS)),
        })
    return Stack(
        'Services',
        template=template,
        TemplateURL=services_template_url,
        Parameters=parameters,
        DependsOn=depends
    )


def add_ecs_cluster(template, depends_on=None):
    """
    Function to add the cluster to the root template.

    :param depends_on: list of dependencies for the object
    :type depends_on: list
    :param template: the root stack template
    :type template: troposphere.Template
    """
    template.add_condition(GENERATED_CLUSTER_NAME_CON_T, GENERATED_CLUSTER_NAME_CON)
    template.add_condition(CLUSTER_NAME_CON_T, CLUSTER_NAME_CON)
    if depends_on is None:
        depends_on = []
    Cluster(
        ROOT_CLUSTER_NAME,
        template=template,
        ClusterName=If(
            CLUSTER_NAME_CON_T,
            Ref('AWS::StackName'),
            Ref(CLUSTER_NAME_T)
        ),
        DependsOn=depends_on
    )


def generate_full_template(session=None, **kwargs):
    """
    Function to generate the root template

    :param session: boto3 session to override client
    :type session: boto3.session.Session

    :return template: Template()
    """
    if session is None:
        session = boto3.session.Session(region_name=kwargs['AwsRegion'])
    template = build_template(
        'Root template generated via ECS ComposeX',
        [USE_FLEET, USE_ONDEMAND, CLUSTER_NAME]
    )
    stack_params = []
    LOG.info(kwargs)
    vpc_stack = None
    compute_stack = None
    if KEYISSET('CreateVpc', kwargs):
        vpc_stack = add_vpc_to_root(template, session, kwargs)
    else:
        generate_vpc_parameters(template, stack_params, **kwargs)
        LOG.info(stack_params)
    if KEYISSET(CLUSTER_NAME_T, kwargs):
        build_parameters_file(stack_params, CLUSTER_NAME_T, kwargs[CLUSTER_NAME_T])
    if KEYISSET('CreateLaunchTemplate', kwargs) or KEYISSET(USE_FLEET_T, kwargs):
        pass
        # if vpc_stack:
        #     compute_stack = add_cluster_to_root(
        #         template, params=stack_params, session=session, vpc_stack=vpc_stack, **kwargs
        #     )
        # else:
        #     compute_stack = add_cluster_to_root(
        #         template, params=stack_params, session=session, **kwargs
        #     )
    depends = add_x_resources(template, session=session, **kwargs)
    if compute_stack:
        depends.append(compute_stack.title)
    if vpc_stack:
        depends.append(vpc_stack.title)
    if KEYISSET('CreateCluster', kwargs):
        add_ecs_cluster(template, depends)
        depends.append(ROOT_CLUSTER_NAME)
    add_services(template, depends, session=session, vpc_stack=vpc_stack, **kwargs)
    with open(kwargs['output_file'], 'w') as root_fd:
        root_fd.write(template.to_json())
    with open(f"{kwargs['output_file'].split('.')[0]}.params.json", 'w') as params_fd:
        params_fd.write(json.dumps(stack_params, indent=4))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
3 Layers subnets calculator for 3-tier VPC
"""

import ipaddress
from math import (
    log,
    ceil
)

pow2_2_prefix = {

    '16': 28,
    '32': 27,
    '64': 26,
    '128': 25,
    '256': 24,
    '512': 23,
    '1024': 22,
    '2048': 21,
    '4096': 20,
    '8192': 19,
    '16384': 18,
    '32768': 17,
    '65536': 16,
    '131072': 15,
    '262144': 14,
    '524288': 13,
    '1048576': 12,
    '2097152': 11,
    '4194304': 10,
    '8388608': 9,
    '16777216': 8
}


def clpow2(x):
    """
    Function to return the closest power of two from given x

    :param x: Number to look the closest power of two for

    :returns: int() closest power of two
    """
    return pow(2, int(log(x, 2) + 0.5))


def nxtpow2(x):
    """Function to find the next power of two from given x number

    :param x: number to look for the next power of two

    :returns: next power of two number
    """
    return int(pow(2, ceil(log(x, 2))))


def cut_per_az(az_cidr, layers_cidr):
    """Subdivide the range per AZ in the region

    :param az_cidr: CIDR to split
    :param layers_cidr: dict() getting updated with layers

    :returns: NIL
    """
    maj_splits = list(az_cidr.subnets(prefixlen_diff=1))
    layers_cidr['app'].append(maj_splits[0])
    min_splits = list(maj_splits[1].subnets(prefixlen_diff=1))
    layers_cidr['pub'].append(min_splits[0])
    layers_cidr['stor'].append(min_splits[1])


def get_subnets(cidr, azs):
    """
    Get the lists of Subnets CIDRs
    """
    cidr = f'{cidr}'
    vpc_net = ipaddress.IPv4Network(cidr)
    number_ips = int(vpc_net.num_addresses - 2)

    if (azs != 2) and (azs % 2):
        azs += 1

    layers_cidr = {
        'app': [],
        'pub': [],
        'stor': []
    }

    ips_per_az = number_ips / azs
    pow2 = clpow2(ips_per_az)
    azs_prefix = pow2_2_prefix['%d' % (pow2)]
    subnets_per_az = list(vpc_net.subnets(new_prefix=azs_prefix))

    for az in subnets_per_az:
        cut_per_az(az, layers_cidr)
    return layers_cidr


def get_subnet_layers(cidr, azs):
    """
    Get Subnets layers based on number of AZs
    """
    layers = get_subnets(cidr, azs)
    cidrs = {}

    for layer in layers:
        cidrs[layer] = []
        sub_list = []
        for subnet in layers[layer]:
            sub_list.append('%s' % (subnet))
        cidrs[layer] = sub_list

    return cidrs


if __name__ == '__main__':
    import json
    from argparse import ArgumentParser
    PARSER = ArgumentParser(description="Splits a CIDR to generate VPC subnets")
    PARSER.add_argument("--cidr", help="CIDR", required=True)
    PARSER.add_argument("--azs", help="AZs", required=True)
    ARGS = PARSER.parse_args()
    CIDRS = get_subnet_layers(ARGS.cidr, int(ARGS.azs))
    print(json.dumps(CIDRS, indent=2))



# -*- coding: utf-8 -*-
"""
Parameters relating to ECS.

This is a crucial part as all the titles, maked `_T` are string which are then used the same way
across all imports, which gives consistency for CFN to use the same names,
which it heavily relies onto.

You can change the names *values* so you like so long as you keep it Alphanumerical [a-zA-Z0-9]
"""

from troposphere import (
    Condition,
    Ref, Equals, And
)

from ecs_composex.ecs import ecs_params
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T

GENERATED_CLUSTER_NAME_CON_T = 'UsCfnGeneratedClusterName'
GENERATED_CLUSTER_NAME_CON = Equals(
    Ref(ecs_params.CLUSTER_NAME),
    ecs_params.CLUSTER_NAME.Default
)

CLUSTER_NAME_CON_T = 'SetClusterNameFromRootStack'
CLUSTER_NAME_CON = And(
    Condition(USE_STACK_NAME_CON_T),
    Condition(GENERATED_CLUSTER_NAME_CON_T)
)
SERVICE_COUNT_ZERO_CON_T = 'ServiceCountIsZeroCondition'
SERVICE_COUNT_ZERO_CON = Equals(
    Ref(ecs_params.SERVICE_COUNT),
    '0'
)

MEM_RES_IS_MEM_ALLOC_CON_T = 'MemoryReservedIsMemoryAllocatedCondition'
MEM_RES_IS_MEM_ALLOC_CON = Equals(
    Ref(ecs_params.MEMORY_RES),
    ecs_params.MEMORY_RES.Default
)

USE_FARGATE_CON_T = 'UseFargateCondition'
USE_FARGATE_CON = Equals(
    Ref(ecs_params.LAUNCH_TYPE),
    'FARGATE'
)

SERVICE_COUNT_ZERO_AND_FARGATE_CON_T = 'ServiceCountZeroAndFargate'
SERVICE_COUNT_ZERO_AND_FARGATE_CON = And(
    Condition(USE_FARGATE_CON_T),
    Condition(SERVICE_COUNT_ZERO_CON_T)
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, Extension, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'troposphere==2.6.0',
    'boto3==1.12.9'
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="John Mille",
    author_email='john@lambda-my-aws.io',
    python_requires='!=2.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Package that generates CFN templates based on a Docker Compose file to generate ECS Cluster,"
                "ECS Services and Extra resources such as SQS Queues, SNS Topics etc. that can be accessed via IAM",
    entry_points={
        'console_scripts': [
            'ecs_composex=ecs_composex.cli:main',
            'ecs_composex-vpc=ecs_composex.vpc.cli:main',
            'ecs_composex-sqs=ecs_composex.sqs.cli:main'
        ]
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='ecs_composex',
    name='ecs_composex',
    packages=find_packages(include=['ecs_composex', 'ecs_composex.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/lambda-my-aws/ecs_composex',
    version='0.1.0',
    zip_safe=False
)



# -*- coding: utf-8 -*-
"""Conditions related to the VPC construction."""

from troposphere import (
    Condition, Equals,
    Ref, Or, Not
)

from ecs_composex.vpc import vpc_params
from ecs_composex.common import cfn_conditions

USE_SUB_ZONE_CON_T = 'UseDnsSubZoneCondition'
USE_SUB_ZONE_CON = Or(
    Equals(
        Ref(vpc_params.USE_SUB_ZONE),
        'True'
    ),
    Condition(cfn_conditions.NOT_USE_CLOUDMAP_CON_T)
)

NOT_USE_VPC_MAP_ID_CON_T = 'NotUseVpcMapId'
NOT_USE_VPC_MAP_ID_CON = Equals(
    Ref(vpc_params.VPC_MAP_ID),
    vpc_params.VPC_MAP_ID.Default
)

USE_VPC_MAP_ID_CON_T = 'UseVpcMapId'
USE_VPC_MAP_ID_CON = Not(Condition(NOT_USE_VPC_MAP_ID_CON_T))



#!/usr/bin/env python
# -*- coding: utf-8 -

"""Console script for ecs_composex.sqs"""
import sys
import argparse

from ecs_composex.common.aws import BUCKET_NAME
from ecs_composex.sqs import create_sqs_template


def main():
    """Console script for ecs_composex."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-b', '--bucket-name', type=str, required=False, default=BUCKET_NAME,
        help='Bucket name to upload the templates to', dest='BucketName'
    )
    parser.add_argument(
        '-f', '--compose-file', required=True, dest="ComposeXFile",
        help="Path to the Docker Compose / ComposeX file"
    )
    parser.add_argument(
        '-o', '--output-file', required=True,
        help="Output file for the template body"
    )
    parser.add_argument('_', nargs='*')
    args = parser.parse_args()

    template = create_sqs_template(**vars(args))
    with open(args.output_file, 'w') as tpl_fd:
        if args.output_file.endswith('.yml') or args.output_file.endswith('.yaml'):
            tpl_fd.write(template.to_yaml())
        else:
            tpl_fd.write(template.to_json())
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-

import boto3
from ecs_composex.common import LOG
from ecs_composex.common import KEYISSET
from ecs_composex.common.aws import get_curated_azs
from ecs_composex.vpc.vpc_template import generate_vpc_template


def create_vpc_template(session=None, **kwargs):
    """Function to create the vpc template for a combined deployment. Invoked by CLI

    :param session:
    :param kwargs:

    :return: vpc_template Template()
    :rtype: troposphere.Template
    """
    azs = []
    if not KEYISSET('AwsAzs', kwargs):
        if KEYISSET('AwsRegion', kwargs):
            azs = get_curated_azs(region=kwargs['AwsRegion'])
        elif session is None:
            session = boto3.session.Session()
            azs = get_curated_azs(session=session)
        else:
            azs = get_curated_azs()

    else:
        azs = kwargs['AwsAzs']
    LOG.debug(azs)
    cidr_block = kwargs['VpcCidr']
    single_nat = KEYISSET('SingleNat', kwargs)
    template = generate_vpc_template(
        cidr_block, azs, session, single_nat=single_nat
    )
    return template



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

import pytest
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from ecs_composex.sqs import create_sqs_template
from ecs_composex.sqs.sqs_perms import generate_sqs_permissions

try:
    BUCKET = os.environ['KNOWN_BUCKET']
except KeyError:
    BUCKET = 'lambda-dev-eu-west-1'


CONFIG = {
    'BucketName': BUCKET,
    'EnvName': 'abcd',
}

HERE = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def content_sqs():
    """
    Opens the file and passes the content around
    """
    with open(f"{HERE}/sqs.yml", 'r') as fd:
        content = yaml.load(fd.read(), Loader=Loader)
    return content

@pytest.fixture
def args():
    return {
        'BucketName': BUCKET,
        'Debug': True
    }


@pytest.fixture
def parser_args():
    return({
        'ComposeXFile': f"{HERE}/services_with_queues.yml",
        'BucketName': BUCKET,
        'OutputFile': '/tmp/sqs.yml'
    })

def test_generate_policies(args):
    """
    Function to test generation of a queue policies
    """
    policies = generate_sqs_permissions(
        'QueueA', {
            'Services': [
                {
                    'name': 'App01',
                    'access': 'RWMessages'
                }
            ]
        },
        **args
    )
    assert len(policies.keys()) == 3


def simulate_cli_input(parser_args):
    """
    Tests
    """
    create_sqs_template(**parser_args)



# -*- coding: utf-8 -*-

"""
Module to create the compute resources, if so chosen, the SpotFleet / OnDemand instances to go with it.

The SpotFleet and OnDemand instances are optional, but the LaunchTemplate gets created so that if
for testing one would wish to run a new EC2 instance, you can simply do it from the launch template.
"""

import boto3

from ecs_composex.common import KEYISSET, load_composex_file
from ecs_composex.common.aws import get_curated_azs
from ecs_composex.compute.cluster_template import generate_cluster_template


def create_cluster_template(session=None, **kwargs):
    """
    Function entrypoint for CLI.
    :param session: boto3 session to override API calls with
    :type session: boto3.session.Session

    :return: cluster template
    :rtype: troposphere.Template
    """
    compose_content = None
    if KEYISSET('ComposeXFile', kwargs):
        compose_content = load_composex_file(kwargs['ComposeXFile'])

    azs = []
    if not KEYISSET('AwsAzs', kwargs):
        if KEYISSET('AwsRegion', kwargs):
            azs = get_curated_azs(region=kwargs['AwsRegion'])
        elif session is None:
            session = boto3.session.Session()
            azs = get_curated_azs(session=session)
        else:
            azs = get_curated_azs()

    else:
        azs = kwargs['AwsAzs']
    return generate_cluster_template(azs, compose_content, **kwargs)



# -*- coding: utf-8 -*-

from troposphere import (
    Condition, Equals, Ref, Not
)
from ecs_composex.compute import cluster_params
USE_SPOT_CON_T = 'UseSpotFleetHostsCondition'
USE_SPOT_CON = Equals(
    Ref(cluster_params.USE_FLEET),
    'True'
)

NOT_USE_SPOT_CON_T = 'NotUseSpotFleetHostsCondition'
NOT_USE_SPOT_CON = Not(
    Condition(USE_SPOT_CON_T)
)

MAX_IS_MIN_T = 'CapacityMaxIsMinCondition'
MAX_IS_MIN = Equals(
    Ref(cluster_params.MAX_CAPACITY),
    0
)



# -*- coding: utf-8 -*-
"""
Functions that will add networking elements for the ECS Service as well as generating
the network configuration for the ServiceDefinition
"""

from troposphere import (
    Ref, Sub, GetAtt, Tags
)
from troposphere.ec2 import (
    SecurityGroup
)
from troposphere.ecs import (
    ServiceRegistry
)
from troposphere.servicediscovery import (
    DnsConfig as SdDnsConfig,
    Service as SdService,
    DnsRecord as SdDnsRecord,
    HealthCheckCustomConfig as SdHealthCheckCustomConfig
)

from ecs_composex.common import KEYISSET, LOG
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.ecs.ecs_loadbalancing import define_service_load_balancing
from ecs_composex.ecs.ecs_params import (
    SERVICE_NAME_T, SERVICE_NAME, SG_T
)
from ecs_composex.vpc.vpc_params import (
    NAMESPACE_ID_IMPORT,
    VPC_ID
)


def add_service_default_sg(template):
    """
    Adds a default security group for the microservice.

    :param template: the template to add the SG to.
    :type template: troposphere.Template

    :return: Security group
    :type: troposphere.ec2.SecurityGroup
    """
    sg = SecurityGroup(
        SG_T,
        template=template,
        GroupDescription=Sub(f"SG for ${{{SERVICE_NAME_T}}} - ${{{ROOT_STACK_NAME_T}}}"),
        Tags=Tags(
            {
                'Name': Sub(f"${{{SERVICE_NAME_T}}}-${{{ROOT_STACK_NAME_T}}}"),
                'StackName': Ref('AWS::StackName'),
                'MicroserviceName': Ref(SERVICE_NAME)
            }
        ),
        VpcId=Ref(VPC_ID)
    )
    return sg


def add_service_to_map(template, service_name, service, settings):
    """
    Function to add the service into a cloudmap
    """
    sd_service = SdService(
        'EcsDiscoveryService',
        template=template,
        Description=f'{service_name}',
        NamespaceId=NAMESPACE_ID_IMPORT,
        HealthCheckCustomConfig=SdHealthCheckCustomConfig(
            FailureThreshold=1.0
        ),
        DnsConfig=SdDnsConfig(
            RoutingPolicy='MULTIVALUE',
            NamespaceId=Ref('AWS::NoValue'),
            DnsRecords=[
                SdDnsRecord(
                    TTL='30',
                    Type='A'
                ),
                SdDnsRecord(
                    TTL='30',
                    Type='SRV'
                )
            ]
        ),
        Name=service_name
    )
    registries = []
    for port in service['ports']:
        registry = ServiceRegistry(
            f"ServiceRegistry{port.split(':')[-1]}",
            RegistryArn=GetAtt(sd_service, 'Arn'),
            Port=port.split(':')[-1]
        )
        registries.append(registry)
    return registries


def define_protocol(port_string):
    """Function to define the port protocol. Default to TCP if not specified otherwise

    :param port_string: the port string to parse from the ports list in the compose file
    :type port_string: str

    :return: protocol, ie. udp or tcp
    :rtype: str
    """
    protocols = ['tcp', 'udp']
    protocol = 'tcp'
    if port_string.find('/'):
        protocol_found = port_string.split('/')[-1].strip()
        if protocol_found in protocols:
            return protocol_found
    return protocol


def define_service_ports(service):
    """Function to define common structure to ports

    :param service: the service as defined in compose file
    :type service: dict

    :return: list of ports the service uses formatted according to dict
    :rtype: list
    """
    ports = []
    valid_keys = ['published', 'target', 'protocol', 'mode']
    service_ports = []
    if KEYISSET('ports', service):
        ports = service['ports']
    for port in ports:
        if isinstance(port, str):
            service_ports.append({
                'protocol': define_protocol(port),
                'published': port.split(':')[0],
                'target': port.split(':')[-1].split('/')[0].strip(),
                'mode': 'awsvpc'
            })
        elif isinstance(port, dict):
            if not set(port).issubset(valid_keys):
                raise KeyError(f"Valid keys are", valid_keys, 'got', port.keys())
            port['mode'] = 'awsvpc'
            service_ports.append(port)
    LOG.debug(service_ports)
    return service_ports


def compile_network_settings(compose_content, service, service_name):
    """Function to generate a set of values used for all network related stuff

    :param compose_content: Docker/ComposeX File content
    :type compose_content: dict
    :param service: service definition as imported from Docker/ComposeX File
    :type service_name: dict
    :param service_name: name of the service as defined in Docker/ComposeX File
    :type service_name: str

    :return: settings
    :rtype: dict
    """
    valid_keys = [
        'use_alb', 'use_nlb', 'use_cloudmap', 'is_public', 'healthcheck', 'ext_sources', 'ports'
    ]
    settings = {}
    globals = {
        'use_cloudmap': True,
        'use_nlb': False,
        'use_alb': True,
        'is_public': False
    }
    if KEYISSET('labels', service):
        settings = service['labels']
    if KEYISSET('configs', compose_content):
        if KEYISSET(service_name, compose_content['configs']):
            LOG.warn(
                f"Service {service_name} has defined configs in the configs section."
                "Overriding with values"
            )
            settings = compose_content['configs'][service_name]
    if (KEYISSET('globals', compose_content['configs']) and
            KEYISSET('network', compose_content['configs']['globals']['network'])):
        globals.update(compose_content['configs']['globals']['network'])
    if not set(settings).issubset(valid_keys):
        LOG.error(valid_keys)
        LOG.error(settings)
        raise KeyError('Invalid keys defined in the network section of the service config in configs')
    for key in valid_keys:
        if key not in settings and key != 'ports':
            if key in globals.keys():
                if key == 'use_cloudmap' and not KEYISSET('is_public', settings):
                    LOG.debug(f'Missing use_cloudmap but {service_name} is public.')
                    settings[key] = False
                else:
                    settings[key] = globals[key]
    settings['ports'] = define_service_ports(service)
    return settings


def define_service_network_config(template, service_name, network_settings, **kwargs):
    """Function to define microservice ingress.

    :param template: service template to add resources to
    :type template: troposphere.Template
    :param service_name: name of the service as defined in the compose file
    :type service_name: str
    :param network_settings: see :func:`ecs_composex.ecs.ecs_networking.compile_network_settings`
    :type network_settings: dict
    :param kwargs: Optional additional parameters

    :return: tuple of the args for the Service() and the dependencies
    :rtype: tuple
    """
    add_service_default_sg(template)
    service_lbs = Ref('AWS::NoValue')
    registries = Ref('AWS::NoValue')
    service_attrs = {
        'LoadBalancers': service_lbs,
        'ServiceRegistries': registries
    }
    external_dependencies = []
    if not KEYISSET('AwsAzs', kwargs):
        raise KeyError(
            'Missing AwsAzs from options.'
            'AZs are required to configure services networking'
        )
    if not KEYISSET('ports', network_settings):
        LOG.debug(f"{service_name} does not have any ports. No ingress necessary")
        return service_attrs, external_dependencies
    LOG.debug(network_settings)
    if KEYISSET('use_nlb', network_settings) or KEYISSET('use_alb', network_settings):
        service_lb = define_service_load_balancing(template, service_name, network_settings, **kwargs)
        service_attrs['LoadBalancers'] = service_lb[0]
        service_attrs['DependsOn'] = service_lb[-1] if isinstance(service_lb[-1], list) else []
    return service_attrs, external_dependencies
